#!/bin/bash

# see the COPYRIGHT.txt file for license information
# Copyright (C) 2016 Ryan Cunningham

# This file is responsible for the creation of a Sugar tablet hard disk
# or hard disk image. AN INTERNET CONNECTION IS REQUIRED.

# Sugar was created by One Laptop per Child to aid in providing com-
# puters to the poor---that is, bridging the `digital divide'. Sugar is
# a very child-friendly interface that turns out to be very conducive to
# tablet use. Sugar is currently developed by Sugar Labs, a member of
# the 501(c)(3) non-profit Software Freedom Conservancy.

# Now, I suspect you should know about the XO Tablet (Sugar originally
# having been produced for the OLPC XO), which was produced to give the
# already-developed world a fair deal, but free and open-source software
# is all about embracing competition.

# Having taught myself about this field, I have been interested in the
# world of Linux programming for years. I am a hetero-sexual adolescent
# White male, but I know and believe that computer programming is not
# just for hetero-sexual adult White or Asian males; the Hour of Code
# and Sugar Labs exist to spread that belief.

# For more information about OLPC, visit <http://one.laptop.org>. For
# more information about Sugar, visit <http://sugarlabs.org>.

# Direct any and all legal concerns about this software to the Software
# Freedom Conservancy.

# This software, specifically, takes one mandatory argument and one
# optional argument. The mandatory argument is the name of the file in
# which to put the installed system. The optional argument is the size
# of that file. The size of that file must follow its name. Names with
# spaces must be in quotation marks. See below for acceptable magnitude
# abbreviations in the file size description.

# LANGUAGE SELECTION IN THE INSTALLED SYSTEM'S INITIAL SETUP PROGRAM
# WILL BE RESTRICTED TO THE LANGUAGE OF THE AREA WHERE THE SYSTEM WAS
# BOUGHT. Language changes, however, may take place after initial setup
# is completed.

# Exit codes:
#        0: Success
#        1: Error creating target image file (virtual installation
#           only)
#        2: Error creating partition table
#        3: Error creating a partition
#        4: Error toggling flags on a partition
#        5: Error creating loopback device file for primary partition
#           (virtual installation only)
#        6: Error making file system
#        7: Error making mount point
#        8: Error mounting file system
#        9: Anaconda installer error
#       10: Error making initial RAM disk
#       11: Error making swap file
#       12: Error formatting swap file
# 13 or 14: Error activating swap file
#       15: Error installing boot loader
#       16: Error configuring boot loader for EFI
#       17: File exists
#       18: Error configuring initial-setup
#       21: Is a directory
#       22: Preconfiguration error

# ---------------------------------------------------------------------

if [ "$1" = "-f" ]
then
    isflashing=yes
    shift
fi
if [ "$1" = "-E" ]
then
    use_efi=yes
    shift
fi

if [ ! "$1" ]
then
    echo "usage: $0 <path> <basearch> [<image-size>]"
    echo "    [-E  use EFI instead of IBM BIOS (useful on i386 only)]"
    echo "    [-f  configure image for flashing to an actual tablet]"
    echo "The suffixes K, G, T, P, E, Z, Y may be used in <image-size>"
    echo "to denote their respective units. At least 5GiB recommended."
    exit 0
fi

# We do not want to overwrite our user's precious content!

if [ -d "$1" ]
then
    echo "$1: (21) Is a directory" >&2
    exit 21
fi
if [ -e "$1" -a ! -b "$1" ]
then
    echo "$1: (17) File exists" >&2
    exit 17
fi

# Do we have a base architecture?

if [ "$2" = "-f" ]
then
    isflashing=yes
    shift
fi
if [ "$2" = "-E" ]
then
    use_efi=yes
    shift
fi

if [ "$2" ]
then
    basearch=$2
else
    echo "A base architecture is needed before installation" >&2
    echo "can continue. It must be 'i386', 'armhfp', or 'x86_64'" >&2
    echo "and must be the same architecture you are using on" >&2
    echo "the host." >&2
    exit 22
fi

# If we do have a third argument, we use that as a file size for
# `fallocate' to parse. If we do not, we use a default size of 32 GiB,
# which is very common for most tablets.

if [ "$3" = "-f" ]
then
    isflashing=yes
    shift
fi
if [ "$3" = "-E" ]
then
    use_efi=yes
    shift
fi

if [ "$3" ]
then
    SIZE=$3
else
    SIZE=32G
fi

# Am I not root?

if [ "$EUID" -ne 0 ]
then
    echo "Sorry, I must be running as the root user." >&2
    exit 22
fi

# We now check for architecture support.

case $basearch in
    i386)
        ;;
    armhfp)
        use_efi=yes;;
    x86_64)
        use_efi=yes;;
    *)
        echo "Sorry, architecture '$basearch' not supported." >&2
        exit 22;;
esac

# If we are using some Linux distribution other than Fedora or some Fedora
# version that's too old, we must now exit.

releasever="$(rpm -q --qf '%{version}\n' fedora-release)"
if [ "$releasever" = "" ]
then
    echo "Sorry, am I not running on Fedora?" >&2
    exit 22
fi
if [ "$releasever" != "rawhide" -a "$releasever" -lt 22 ]
then
    echo "Sorry, your Fedora is too old." >&2
    echo "I require Fedora 22 or later." >&2
    exit 22
fi

# If we are installing a Fedora version that doesn't exist, or our user is
# not connected to the Internet, we must exit now.

if [ "$releasever" != "rawhide" ] && \
  ! curl -f http://archive.fedoraproject.org/pub/fedora/linux/releases/\
$releasever/Everything/$basearch/os/ >/dev/null 2>/dev/null
then
    echo "Sorry, Fedora has not published version '$releasever'" >&2
    echo "for production on '$basearch', there is no Fedora" >&2
    echo "'$releasever', or you don't have an Internet" >&2
    echo "connection." >&2
    exit 22
fi

# We can't run more than one concurrent instance of ourselves!

if [ -f /var/lock/create-image.lock ]
then
    echo "$0 already running with pid $(cat /var/lock/create-image.lock)" >&2
    exit 22
fi
echo $$ > /var/lock/create-image.lock

# Now we specify what to do when we exit from now on.

if [ "$use_efi" = yes ]
then
    DEV=${1}2
else
    DEV=${1}1
fi
remove_files() {
    umount -fl /mnt/sysimage/dev/*
    umount -fl /mnt/sysimage/*
    if [ -b "$DEV" ]
    then
        umount -fl $DEV
    else
        umount -fl /dev/loop0
        losetup -d /dev/loop0
    fi
    rm fedora-repo-not-rawhide.ks \
       fedora-repo-rawhide.ks \
       linux_path_temp \
       initrd_path_temp \
       fedora-live-base.ks \
       fedora-live-minimization.ks \
       fedora-live-soas.ks \
       fedora-repo.ks \
       fedora-soas-packages.ks \
       /var/lock/create-image.lock
}
exit_trap() {
    remove_files
    if [ "$1" ]
    then
        exit $1
    else
        exit 0
    fi
    # NOT REACHED
}
trap "exit_trap" INT HUP ABRT TERM

# Here we check if all the programs and auxiliary files we need are installed.

echo "Installing program dependencies ..."
./install-or-update-deps.sh

# We now create a disk image (or use a real disk) and create a partition with
# the given size.

if [[ "$SIZE" != *[[:alpha:]]* ]]
then
    echo "Internal error: Disk size specification must contain a suffix" >&2
    exit 22
fi
if [ ! -b "$1" ]
then
    echo "Creating $SIZE image at $1 ..."
    fallocate -l $SIZE $1 || exit_trap 1
else
    echo "Formatting disk $1 with size $SIZE ..."
fi
parted $1 mktable msdos || exit_trap 2

if [ "$use_efi" = yes ]
then
    # Create the EFI System Partition.
    parted $1 mkpart boot ext2 2048s 3M || exit_trap 3
    parted $1 toggle 1 esp || exit_trap 4
    ROOTNUM=2
else
    ROOTNUM=1
fi

# Create the root partition.

parted $1 mkpart primary ext4 3M $SIZE || exit_trap 3
parted $1 toggle $ROOTNUM boot || exit_trap 4
partprobe

# And we now give the partition a file system.

echo "Formatting root partition on $1 with ext4 file system ..."
if [ -b "$DEV" ]
then
    umount -fl $DEV
else
    umount -fl /dev/loop0
    losetup -d /dev/loop0
fi
if [ -b "$1" ]
then
    if [ "$use_efi" = yes ]
    then
        DEVPATH="${1}2"
    else
        DEVPATH="${1}1"
    fi
    mkfs.ext4 $DEVPATH || exit_trap 6
else
    # We can safely mount the root at 3145728 bytes even if there's an EFI
    # System Partition before then.
    losetup -o 3145728 /dev/loop0 $1 || exit_trap 5
    mkfs.ext4 /dev/loop0 || exit_trap 6
    DEVPATH=/dev/loop0
fi

# The partition must be mounted before Anaconda can install to it. We mount the
# partition now.

if [ -b "$1" ]
then
    echo "WARNING: Please do not disconnect the target disk"
    echo "while installation is in progress."
fi
echo "Mounting $1 ..."
mkdir -p /mnt/sysimage || exit_trap 7
mount -t ext4 $DEVPATH /mnt/sysimage || exit_trap 8

# Now we go ahead and substitute the release version and architecture where
# appropriate and then run Anaconda. This will take up most of the necessary
# time---perhaps fifteen minutes at maximum.

echo "Installing the base system ..."
sed -e "s/{RELEASEVER}/$releasever/g" -e "s/{BASEARCH}/$basearch/g" \
    fedora-repo-not-rawhide.ks.tmpl > fedora-repo-not-rawhide.ks
sed -e "s/{BASEARCH}/$basearch/g" fedora-repo-rawhide.ks.tmpl \
    > fedora-repo-rawhide.ks
python3 -u /usr/sbin/anaconda -C --kickstart fedora-live-soas.ks --dirinstall ||
    exit_trap 9
patch -p0 -d/ -i $(pwd)/firstboot-windowmanager.patch
patch -p0 -d/ -i $(pwd)/bashrc.patch

# Here are some things Anaconda can't do with this type of installation. Because
# of that, we do them ourselves.

echo "Generating initramfs ..."

# First we create an initial RAM file system. (Anaconda tries to create one but
# it is deleted by a post-installation script.) These are necessary for Fedora's
# kernel to boot the tablet.

find /mnt/sysimage/boot -name vmlinuz-\* ! -name vmlinuz-0-rescue\* \
    ! -name vmlinuz\*+PAE\* | sed 's@/mnt/sysimage/@/@g' |
    head -1 > linux_path_temp
linux_path="$(cat linux_path_temp)"
# 'uname -r' reports the host's kernel version; the kernel version may be
# different in the guest. For that reason, we cannot use 'uname -r'.
kernversion="$(echo $linux_path | sed 's@/boot/vmlinuz-@@g')"
systemd-nspawn -D /mnt/sysimage bash -c \
    "dracut --no-hostonly --kver $kernversion" || exit_trap 10
systemd-nspawn -D /mnt/sysimage bash -c \
    "dracut --no-hostonly --kver ${kernversion}+PAE" || exit_trap 10
systemd-nspawn -D /mnt/sysimage bash \
    /etc/kernel/postinst.d/51-dracut-rescue-postinst.sh "$kernversion" \
    "$linux_path" || exit_trap 10

# Now we create a swap image. A swap file is desired here (as opposed to a swap
# partition), because a swap file is easier from an installation script.

echo "Generating swap space ..."
fallocate -l 1G /mnt/sysimage/boot/swap.img || exit_trap 11
mkswap /mnt/sysimage/boot/swap.img || exit_trap 12

# Next we configure the file system table to represent what we have now done.

echo "Configuring file systems ..."
if [ "$isflashing" = yes ]
then
    if [ "$use_efi" = "yes" ]
    then
        fspath=/dev/sda2
    else
        fspath=/dev/sda1
    fi
elif [ -b $1 ]
then
    if [ "$use_efi" = "yes" ]
    then
        fspath=${1}2
    else
        fspath=${1}1
    fi
elif [ "$use_efi" = "yes" ]
then
    fspath=/dev/sda2
else
    fspath=/dev/sda1
fi
cat > /mnt/sysimage/etc/fstab.new << EOF || exit_trap 13
$fspath / ext4 defaults 1 1
$(cat /mnt/sysimage/etc/fstab)
/boot/swap.img swap swap defaults 0 0
EOF
mv /mnt/sysimage/etc/fstab.new /mnt/sysimage/etc/fstab || exit_trap 14

# And now we install the Grand Unified Bootloader using the configuration file
# template in this directory.

echo "Installing and configuring boot loader ..."
if [ "$use_efi" = "yes" ]
then
    root="hd0,msdos2"
else
    root="hd0,msdos1"
fi
cp grub.cfg.tmpl /mnt/sysimage/boot/grub2
find /mnt/sysimage/boot -name initramfs\* \
    ! -name initramfs-0-rescue\* ! -name initramfs\*+PAE\* |
    sed 's@/mnt/sysimage@/@g' | head -1 > initrd_path_temp
initrd_path="$(cat initrd_path_temp)"
sed -e "s@%%LINUX_PATH%%@$linux_path@g" -e "s@%%INITRD_PATH%%@$initrd_path@g" \
    -e "s@%%FS_PATH%%@$fspath@g" -e "s@%%ROOT%%@($root)@g" \
    /mnt/sysimage/boot/grub2/grub.cfg.tmpl > /mnt/sysimage/boot/grub2/grub.cfg
rm /mnt/sysimage/boot/grub2/grub.cfg.tmpl
if [ "$basearch" = "i386" -a "$use_efi" = "yes" ]
then
    grub_target=i386-efi
elif [ "$basearch" = "i386" ]
then
    grub_target=i386-pc
elif [ "$basearch" = "armhfp" ]
then
    grub_target=arm-efi
elif [ "$basearch" = "x86_64" ]
then
    grub_target=x86_64-efi
fi
grub2-install --force --boot-directory=/mnt/sysimage/boot --target=$grub_target \
    $1 || exit_trap 15
if [ "$use_efi" = "yes" ]
then
    (
        mkdir /mnt/efi
        if [ -b $1 ]
        then
            mount "${1}1" /mnt/efi
        else
            losetup -o 1048576 /dev/loop1 $1
            mount /dev/loop1 /mnt/efi
        fi
        cp /mnt/sysimage/boot/grub2/grub.cfg /mnt/efi/EFI/fedora/grub.cfg
        umount -fl /mnt/efi
        if [ ! -b $1 ]
        then
            losetup -d /dev/loop1
        fi
    ) || exit_trap 16
fi

# Now, we configure initial-setup to force it to display the language panel,
# which, if not displayed, may give users who do not speak English a sense of
# racism.

echo "Configuring initial-setup ..."
sed -i -e '/lang/d' -e '/keyboard/d' \
    -e 's/^firstboot.*/firstboot --enable --reconfig/g' \
    /mnt/sysimage/root/anaconda-ks.cfg || exit_trap 18
echo 'user --groups=wheel --name=tablet --gecos=Tablet' \
    >> /mnt/sysimage/root/anaconda-ks.cfg

# And now we clean up and exit.

echo "Done."
exit_trap 0
